From 60bd01b8752def70685cfb67fe47bf0ae9be3f02 Mon Sep 17 00:00:00 2001
From: "Matwey V. Kornilov" <matwey.kornilov@gmail.com>
Date: Thu, 31 Jan 2019 13:20:40 +0300
Subject: [PATCH v2 3/3] Add initial support for spgist quadtree
 @<(point,circle) operator

Signed-off-by: Matwey V. Kornilov <matwey.kornilov@gmail.com>
---
 src/backend/access/spgist/spgquadtreeproc.c       | 73 +++++++++++++++--
 src/include/catalog/pg_amop.dat                   |  3 +
 src/test/regress/expected/create_index_spgist.out | 96 +++++++++++++++++++++++
 src/test/regress/sql/create_index_spgist.sql      | 32 ++++++++
 4 files changed, 197 insertions(+), 7 deletions(-)

diff --git a/src/backend/access/spgist/spgquadtreeproc.c b/src/backend/access/spgist/spgquadtreeproc.c
index 41e7fe5c81..4dd2246a92 100644
--- a/src/backend/access/spgist/spgquadtreeproc.c
+++ b/src/backend/access/spgist/spgquadtreeproc.c
@@ -143,6 +143,44 @@ spg_quad_inner_consistent_box_helper(ScanKey sk, Point *centroid)
 	return r;
 }
 
+static int
+spg_quad_inner_consistent_circle_helper(ScanKey sk, Point *centroid)
+{
+	/*
+	 * Evaluate distances between the query center point and the quadrants.
+	 * The distance between a set and a point is the minimal possible distance
+	 * between any set point and the point. There are three possible cases: the
+	 * point belongs to the quandrand, the point projection is on the quadrant
+	 * edge, the point projections is on the quadrant vertex.
+	 */
+	CIRCLE *circleQuery = DatumGetCircleP(sk->sk_argument);
+	int r = 0;
+
+	const Point cp = {
+		.x = float8_mi(centroid->x, circleQuery->center.x),
+		.y = float8_mi(centroid->y, circleQuery->center.y)
+	};
+
+	const float8 x_p0 = (0. > cp.x ? 0. : cp.x);
+	const float8 y_p0 = (0. > cp.y ? 0. : cp.y);
+	const float8 x_n0 = (0. < cp.x ? 0. : cp.x);
+	const float8 y_n0 = (0. < cp.y ? 0. : cp.y);
+
+	const float8 d[4] = {
+		HYPOT(x_p0, y_p0),
+		HYPOT(x_p0, y_n0),
+		HYPOT(x_n0, y_n0),
+		HYPOT(x_n0, y_p0)
+	};
+
+	for (int i = 0; i < 4; i++) {
+		if (d[i] <= circleQuery->radius)
+			r |= (1 << (i + 1));
+	}
+
+	return r;
+}
+
 Datum
 spg_quad_choose(PG_FUNCTION_ARGS)
 {
@@ -356,7 +394,18 @@ spg_quad_inner_consistent(PG_FUNCTION_ARGS)
 					which &= (1 << 1) | (1 << 4);
 				break;
 			case RTContainedByStrategyNumber:
-				which &= spg_quad_inner_consistent_box_helper(sk, centroid);
+
+				switch (sk->sk_subtype) {
+				case BOXOID:
+					which &= spg_quad_inner_consistent_box_helper(sk, centroid);
+					break;
+				case CIRCLEOID:
+					which &= spg_quad_inner_consistent_circle_helper(sk, centroid);
+					break;
+				default:
+					elog(ERROR, "unrecognized right type OID: %d", sk->sk_subtype);
+					break;
+				}
 				break;
 			default:
 				elog(ERROR, "unrecognized strategy number: %d", sk->sk_strategy);
@@ -443,12 +492,22 @@ spg_quad_leaf_consistent(PG_FUNCTION_ARGS)
 				break;
 			case RTContainedByStrategyNumber:
 
-				/*
-				 * For this operator, the query is a box not a point.  We
-				 * cheat to the extent of assuming that DatumGetPointP won't
-				 * do anything that would be bad for a pointer-to-box.
-				 */
-				res = SPTEST(box_contain_pt, query, datum);
+				switch (sk->sk_subtype) {
+				case BOXOID:
+					/*
+					 * For this operator, the query is a box not a point.  We
+					 * cheat to the extent of assuming that DatumGetPointP won't
+					 * do anything that would be bad for a pointer-to-box.
+					 */
+					res = SPTEST(box_contain_pt, query, datum);
+					break;
+				case CIRCLEOID:
+					res = SPTEST(circle_contain_pt, query, datum);
+					break;
+				default:
+					elog(ERROR, "unrecognized right type OID: %d", sk->sk_subtype);
+					break;
+				}
 				break;
 			default:
 				elog(ERROR, "unrecognized strategy number: %d", sk->sk_strategy);
diff --git a/src/include/catalog/pg_amop.dat b/src/include/catalog/pg_amop.dat
index cf63eb7d54..6ea7e21aa1 100644
--- a/src/include/catalog/pg_amop.dat
+++ b/src/include/catalog/pg_amop.dat
@@ -1373,6 +1373,9 @@
   amoprighttype => 'box', amopstrategy => '8', amopopr => '<@(point,box)',
   amopmethod => 'spgist' },
 { amopfamily => 'spgist/quad_point_ops', amoplefttype => 'point',
+  amoprighttype => 'circle', amopstrategy => '8', amopopr => '<@(point,circle)',
+  amopmethod => 'spgist' },
+{ amopfamily => 'spgist/quad_point_ops', amoplefttype => 'point',
   amoprighttype => 'point', amopstrategy => '15', amoppurpose => 'o',
   amopopr => '<->(point,point)', amopmethod => 'spgist',
   amopsortfamily => 'btree/float_ops' },
diff --git a/src/test/regress/expected/create_index_spgist.out b/src/test/regress/expected/create_index_spgist.out
index 81b4a67c39..191e3088af 100644
--- a/src/test/regress/expected/create_index_spgist.out
+++ b/src/test/regress/expected/create_index_spgist.out
@@ -50,6 +50,102 @@ SELECT count(*) FROM quad_point_tbl WHERE box '(200,200,1000,1000)' @> p;
   1057
 (1 row)
 
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,399),1.41>';
+ count 
+-------
+     0
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,399),1.42>';
+ count 
+-------
+  1000
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,401),1.41>';
+ count 
+-------
+     0
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,401),1.42>';
+ count 
+-------
+  1000
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,399),1.41>';
+ count 
+-------
+     0
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,399),1.42>';
+ count 
+-------
+  1000
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,401),1.41>';
+ count 
+-------
+     0
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,401),1.42>';
+ count 
+-------
+  1000
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(333,399),0.99>';
+ count 
+-------
+     0
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(333,399),1.01>';
+ count 
+-------
+  1000
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(333,401),0.99>';
+ count 
+-------
+     0
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(333,401),1.01>';
+ count 
+-------
+  1000
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,400),0.99>';
+ count 
+-------
+     0
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,400),1.01>';
+ count 
+-------
+  1000
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,400),0.99>';
+ count 
+-------
+     0
+(1 row)
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,400),1.01>';
+ count 
+-------
+  1000
+(1 row)
+
 SELECT count(*) FROM quad_point_tbl WHERE p << '(5000, 4000)';
  count 
 -------
diff --git a/src/test/regress/sql/create_index_spgist.sql b/src/test/regress/sql/create_index_spgist.sql
index 8e6c453307..25881387c3 100644
--- a/src/test/regress/sql/create_index_spgist.sql
+++ b/src/test/regress/sql/create_index_spgist.sql
@@ -42,6 +42,38 @@ SELECT count(*) FROM quad_point_tbl WHERE p <@ box '(200,200,1000,1000)';
 
 SELECT count(*) FROM quad_point_tbl WHERE box '(200,200,1000,1000)' @> p;
 
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,399),1.41>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,399),1.42>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,401),1.41>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,401),1.42>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,399),1.41>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,399),1.42>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,401),1.41>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,401),1.42>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(333,399),0.99>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(333,399),1.01>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(333,401),0.99>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(333,401),1.01>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,400),0.99>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(332,400),1.01>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,400),0.99>';
+
+SELECT count(*) FROM quad_point_tbl WHERE p <@ circle '<(334,400),1.01>';
+
 SELECT count(*) FROM quad_point_tbl WHERE p << '(5000, 4000)';
 
 SELECT count(*) FROM quad_point_tbl WHERE p >> '(5000, 4000)';
-- 
2.13.7

